/*
* data remapping primitive
*/
#include <stdio.h>
#include "parti.h"

#define MAP_MSG  10000

/******************************************************************/
/* 							          */
/* given a distributed processor assignment array(color),         */
/* redistribute the initial mapping (cur_local_ind)               */
/* and return a  schedule (sched) to remap any array that might   */
/* be aligned to this particular initial mapping(cur_local_ind)   */
/*                                                                */
/******************************************************************/

remap_old(color, cur_local_ind, cur_n_data,  sched, new_local_ind, newndata)
int *color, *cur_local_ind;
int cur_n_data;
SCHED **sched;
int *new_local_ind, *newndata;
{ 
 int i, j, no;
 int node;
 int *pcount, *bcount, psize;
 int *buffer, **buckets;
 int *wbuf;
 int recv_n;
 int proc, bsize, isize;
 int *Local, *Proc;
 int *msgid;
 TTABLE *otable;
 int intcompare();
 int remap_n_data;

 
 char fname[10];
 FILE *fp;

 otable  = build_translation_table(BLOCKED, cur_local_ind, cur_n_data);

 buckets = (int **) malloc(sizeof(int *)*PARTI_numnodes);
 pcount  = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 bcount  = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 wbuf    = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 buffer  = (int *) malloc(sizeof(int)*(cur_n_data));
 
 for(i=0; i<PARTI_numnodes; i++) pcount[i] = 0;

 for(i=0; i<cur_n_data; i++) { 
     pcount[color[i]]++;
 }

 psize = 0;
 for(i=0; i<PARTI_numnodes; i++) {
   buckets[i] = &buffer[psize];
   psize    += pcount[i];
 }

 for(i=0; i<PARTI_numnodes; i++) bcount[i] = 0;

 for(i=0; i<cur_n_data; i++) {
   proc = color[i];
   buckets[proc][bcount[proc]++] = cur_local_ind[i];
 }

 PARTI_gisum(pcount, PARTI_numnodes, wbuf);

 recv_n = pcount[PARTI_mynode];

 msgid = wbuffer->msg_id;
 for(i = 0, no = 0; i < PARTI_numnodes; i++) {
   if (pcount[i] > 0) {
     msgid[no++] = PARTI_isend(MAP_MSG, buckets[i], 
				sizeof(int)*bcount[i], i, PARTI_pid); 
   }
 } 

 remap_n_data = recv_n;

 psize = 0;
 while (recv_n > 0) {
#ifdef	SP1
   node = DONTCARE;
   bsize = recv_n * sizeof(int);
#else
   PARTI_cprobe(MAP_MSG);
   bsize = PARTI_infocount();
   node  = PARTI_infonode();
#endif
   PARTI_crecv(MAP_MSG, &new_local_ind[psize], bsize, node, pid);
#ifdef	SP1
   bsize = PARTI_infocount();
#endif
   isize = bsize/sizeof(int);
   psize  += isize;
   recv_n -= isize;
 }
 
 for (i = 0; i < no; i++) {
   PARTI_msgwait(msgid[i]);
 }
 
 free(buckets); free(pcount); free(buffer); free(wbuf); free(bcount);

 Local  = (int *) malloc(sizeof(int)*(remap_n_data));
 Proc   = (int *) malloc(sizeof(int)*(remap_n_data));

 qsort(new_local_ind, remap_n_data, sizeof(int), intcompare);

 dereference(otable, new_local_ind, Local, Proc, remap_n_data);
 *sched = schedule(Local, Proc, remap_n_data, 1);
 *newndata = remap_n_data;

 free(Local); free(Proc); 
 free_table(otable);

}

static int
intcompare(x, y)
int *x, *y;
{
 return (*x-*y);
}


/******************************************************************/
/* 							          */
/* given a distributed processor assignment distributed table(table),*/
/* redistribute the initial mapping (cur_local_ind)               */
/* and return a  schedule (sched) to remap any array that might   */
/* be aligned to this particular initial mapping(cur_local_ind)   */
/*                                                                */
/******************************************************************/

remap(table, cur_local_ind,  sched, new_local_ind, newndata)
TTABLE *table;
int  *cur_local_ind;
SCHED **sched;
int *new_local_ind, *newndata;
{ 
 int i, j, no;
 int node;
 int *pcount, *bcount, psize;
 int *buffer, **buckets;
 int *wbuf;
 int recv_n;
 int proc, bsize, isize;
 int *Local, *Proc;
 int *color;
 int *msgid;
 TTABLE *otable;
 int intcompare();
 int remap_n_data;
 int cur_n_data;

 
 char fname[10];
 FILE *fp;

 color      = (int *)trans_table_color(  table );
 cur_n_data = trans_table_size( table );

 otable  = build_translation_table(BLOCKED, cur_local_ind, cur_n_data);

 buckets = (int **) malloc(sizeof(int *)*PARTI_numnodes);
 pcount  = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 bcount  = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 wbuf    = (int *) malloc(sizeof(int *)*PARTI_numnodes);
 buffer  = (int *) malloc(sizeof(int)*(cur_n_data));
 
 for(i=0; i<PARTI_numnodes; i++) pcount[i] = 0;

 for(i=0; i<cur_n_data; i++) { 
     pcount[color[i]]++;
 }

 psize = 0;
 for(i=0; i<PARTI_numnodes; i++) {
   buckets[i] = &buffer[psize];
   psize    += pcount[i];
 }

 for(i=0; i<PARTI_numnodes; i++) bcount[i] = 0;

 for(i=0; i<cur_n_data; i++) {
   proc = color[i];
   buckets[proc][bcount[proc]++] = cur_local_ind[i];
 }

 PARTI_gisum(pcount, PARTI_numnodes, wbuf);

 recv_n = pcount[PARTI_mynode];

 msgid = wbuffer->msg_id;
 for(i = 0, no = 0; i < PARTI_numnodes; i++) {
   if (bcount[i] > 0) {
     msgid[no++] = PARTI_isend(MAP_MSG, buckets[i], 
				sizeof(int)*bcount[i], i, PARTI_pid); 
   }
 } 

 remap_n_data = recv_n;

 psize = 0;
 while (recv_n > 0) {
#ifdef  SP1
   node = DONTCARE;
   bsize = recv_n * sizeof(int);
#else
   PARTI_cprobe(MAP_MSG);
   bsize = PARTI_infocount();
   node  = PARTI_infonode();
#endif
   PARTI_crecv(MAP_MSG, &new_local_ind[psize], bsize, node, pid);
#ifdef  SP1
   bsize = PARTI_infocount();
#endif
   isize = bsize/sizeof(int);
   psize  += isize;
   recv_n -= isize;
 }

 for (i = 0; i < no; i++) {
   PARTI_msgwait(msgid[i]);
 }
 
 free(buckets); free(pcount); free(buffer); free(wbuf); free(bcount);

 Local  = (int *) malloc(sizeof(int)*(remap_n_data? remap_n_data: 1));
 Proc   = (int *) malloc(sizeof(int)*(remap_n_data? remap_n_data: 1));

 qsort(new_local_ind, remap_n_data, sizeof(int), intcompare);

 dereference(otable, new_local_ind, Local, Proc, remap_n_data);
 *sched = schedule(Local, Proc, remap_n_data, 1);
 *newndata = remap_n_data;

 free(Local); free(Proc); 
 free_table(otable);
}
